#
/*
 *	    M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *			   Douglas P. Kingston III
 *
 *			       Batch Job Queuer
 *
 *	________________________________________________________________
 *
 *				Revision History
 *				================
 *	21 Sep 82  DPK	Modified to use getdate(3).
 *	10 Dec 82  DPK	Corrected handling of copy() with stdin.
 *	20 Dec 82  DPK	Elinimated the fatal() function.
 */

#include <stdio.h>
#include <pwd.h>
#include <signal.h>
#include "queue.h"

#ifndef V6
extern	char	**environ;
#endif
extern	int	Pri_min, Pri_max;
extern	char	*Tmpqdir;


/*
 *	Configurations Values for BATCH
 */
int	Batchpri	= 64;		/* Default priority (in the middle) */
char	*Batchforms	= "Shell";	/* Dummy */
char	*Batchqueue	= "batch";	/* Default queue for batch jobs */
char	*Sysmgr		= "mdqs";	/* Default orphanage for messages */
char	*Hostname	= "localhost";	/* Default local host name */

extern struct passwd *getpwuid();
extern char	*getenv();
extern char	*index();
extern char	*rindex();
extern char	**findque();
extern FILE	*popen();
extern time_t	getdate();
extern int	cleanup();
extern long	copy();

char	*queue		= NULL;
char	*shell		= NULL;
char	*output		= NULL;
char	*title		= NULL;		/* Job title */
char	*username	= NULL;		/* Internet Return Address */
char	missingfmt[]	= "Missing %s argument";
char	dupfmt[]	= "Duplicate %s argument given";

int	priority	= 0;
int	boptions	= 0;
int	qoptions	= 0;
int	verbose		= 0;
int	error		= 0;
int	origumask	= 0;
time_t	atime		= 0L;
struct	passwd *pw;

struct	request r;
FILE	*cf;			/* The control file */
long	timenow;

FILE	*mkdata();
FILE	*mkcontrol();

struct paramtab paramtab[] = {
	"sysmgr",	STRPVAR,	&Sysmgr,		NULL,
	"hostname",	STRPVAR,	&Hostname,		NULL,
	"batch-forms",	STRPVAR,	&Batchforms,		NULL,
	"batch-queue",	STRPVAR,	&Batchqueue,		NULL,
	"batch-prior",	INTVAR,		(char **) &Batchpri,	NULL,
	0,		0,		0,			0
};

/*
 *		M A I N ( )
 */
main (argc, argv)
int	argc;
char	**argv;
{
	FILE	*df;			/* Data file, shell stdin */
	FILE	*fin;
	char	dataname[QDIRSIZ];
	char	*arg;
	char	*cp;
	char	**qp;
	int	i;

	/*  We only act with the privileges of the REAL id  */
	r.r_gid = getgid();
	r.r_uid = getuid();
	time (&timenow);

	signal( SIGHUP, cleanup );
	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
		(void) signal( SIGQUIT, cleanup );
	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
		(void) signal( SIGINT, cleanup );

	
	rdconf( paramtab );

	r.r_copies = 1;
	r.r_pri = Batchpri;
	r.r_tstart = 0L;
	for (i = 1; i < argc; i++) {
		if (*argv[i] != '-')
			break;
		arg = argv[i];
		arg++;
		while (*arg) {
			switch (*arg) {
			case 'a':	/* Run after time */
				if (argv[i+1] == 0) {
					err (missingfmt, "time");
					error++;
				} else
					i++;
				if (atime != 0L) {
					err (dupfmt, "time");
					error++;
				} else if ((atime = getdate(argv[i], NULL)) == -1L) {
					err ("Bad time specification");
					error++;
				} else
					r.r_tstart = atime;
				break;

			case 'm':	/* Send mail upon completion */
				qoptions |= O_MAIL;
				break;

			case 'o':
				if (argv[i+1] == 0) {
					err (missingfmt, "output");
					error++;
				} else
					i++;
				if (output != 0) {
					err (dupfmt, "output");
					error++;
				}
				output = argv[i];
				break;

			case 'p':
				if (argv[i+1] == 0) {
					err (missingfmt, "priority");
					error++;
				} else
					i++;
				if (priority != 0) {
					err (dupfmt, "priority");
					error++;
				} else if (checknum(argv[i],
				    &priority, Pri_min, Pri_max) == -1) {
					fprintf (stderr,
				  "Priorities must be in the range %d to %d",
					  Pri_min, Pri_max);
					error++;
				} else
					r.r_pri = priority;
				break;

			case 'q':
				if (argv[i+1] == 0) {
					err (missingfmt, "queue");
					error++;
				} else
					i++;
				if (queue != 0) {
					err (dupfmt, "queue");
					error++;
				} else if((qp = findque( argv[i] )) == NULL) {
					err ("Unknown queue '%s'", argv[i]);
					error++;
				}
				queue = argv[i];
				break;

			case 's':
				if (argv[i+1] == 0) {
					err (missingfmt, "shell");
					error++;
				} else
					i++;
				if (shell != 0) {
					err (dupfmt, "shell");
					error++;
				}
				shell = argv[i];
				break;

			case 't':
				if (argv[i+1] == 0) {
					err (missingfmt, "title");
					error++;
				} else
					i++;
				if (title != 0) {
					err (dupfmt, "title");
					error++;
				}
				title = argv[i];
				break;

			case 'u':
				if (argv[i+1] == 0) {
					err (missingfmt, "user");
					error++;
				} else
					i++;
				if (username != 0) {
					err (dupfmt, "user");
					error++;
				}
				username = argv[i];
				break;

			case 'v':
				verbose++;
				break;

			case '?':
				error++;
				break;

			default:
				err ("Unknown flag %c", *arg);
				error++;
			}
			arg++;
		}
	}
	
	if( argc - i > 1 ) {
		err( "Too many arguments" );
		error++;
	}

	if( queue == 0 )
		if( findque( queue = Batchqueue ) == NULL )
			bomb( "Default queue '%s' not configured\n", queue );

	if( error ) {
		fputs( "Usage:   batch [options] [shellfile]\n", stderr );
		fputs( "Options: [-a time] [-m] [-o file] [-p pri] [-q que]\n",  stderr );
		fputs( "         [-s shell] [-t title] [-u useraddess] [-v]\n", stderr );
		exit (1);
	}

	origumask = umask (077);

	cf = mkcontrol ("QbaXXXXXX", r.r_name);
	r.r_options = qoptions;
	strncpy (r.r_que, queue, QUESIZ);
	strncpy (r.r_forms, Batchforms, FORMSIZ);

	/* Write out the request structure */
	puthdr (&r, cf);

	/* Write out the one line INTERNET address for notification */
	if( username == 0 ) {
		if ((pw = getpwuid (r.r_uid)) == NULL)
			fprintf( cf, "UUser Unknown <%s>\n", Sysmgr );
		else
			fprintf( cf, "U%s@%s\n", pw->pw_name, Hostname );
	} else
		fprintf( cf, "U%s\n", username );

	/* Write out job title comment */
	if( title )
		fprintf( cf, "T%s\n", title );


	/* If special BATCH options were specified, note them in DECIMAL */
	if (boptions)
		fprintf (cf, "O%d\n", boptions);

	if( shell == 0 )
		shell = getenv( "SHELL" );

	if( shell )
		fprintf( cf, "S%s\n", shell );
	if( output )
		fprintf( cf, "P%s\n", output );

	/* Store the environment away */
	snapshot( cf );

	/*
	 *  Read shell commands from the standard input.
	 */
	df = mkdata (timenow++, dataname);
	if( argc == i ) {
		long	total;

		/* Process the standard input */
		fin = stdin;
		total = copy( stdin, df, "STDIN" );
		fprintf (cf, "TSTDIN\nC%ld\n", total);
	} else {
		long	total;

		/* Shell commands are in a specified file */
		if (access (argv[i], RACCESS))
			bomb ("Cannot access %s\n", argv[i]);
		if ((fin = fopen (argv[i], "r")) == NULL)
			bomb( "Cannot open %s (%s)\n", argv[i],
				sys_errlist[errno] );
		total = copy (fin, df, argv[i]);
		if (cp = rindex (argv[i], '/'))
			cp++;
		else
			cp = argv[i];
		fprintf (cf, "T%s\nC%ld\n", cp, total);
	}
	fclose (df);
	fprintf (cf, "I%s\n", dataname);
	fclose (cf);
	if( ferror( cf ) || ferror( df ) ) {
		err( "Error creating queue files, aborting" );
		exit( 99 );
	}
	queueit(&r, verbose);

	/**** Never Reached ****/
}

/*
 *		S N A P S H O T ( )
 *  This function tries to record as much of the runtime
 *  environment that it can at the time batch was invoked.
 *  among the items of interest are the current directory
 *  and the environment variables.
 */
snapshot( cf )
FILE	*cf;
{
	char	tmpbuf[PATHSIZ];
	register char	**cpp;
	register char	*cp;
	FILE	*pf;

	dumpdir( cf );

#ifndef V6
	/*
	 *	Now swallow up the environment
	 */
	for( cpp = environ; *cpp != NULL; cpp++ ) {
		cp = index( *cpp, '\n' );
		if( cp != NULL )
			*cp = 0;
		fprintf( cf, "E%s\n", *cpp );
	}
#endif
	fprintf( cf, "M%d\n", origumask);
}

cleanup(signal)
int signal;
{
	rmdata( cf );
	if( r.r_name[0] ) {
		char	longname[PATHSIZ];
	
		sprintf (longname, "%s/%s", Tmpqdir, r.r_name);
		unlink (longname);
	}
	_exit( signal );
}
