/*
 *			D M A I N . C 
 *
 * $Revision: 1.10 $
 *
 * $Log:	dmain.c,v $
 * Revision 1.10  89/02/23  13:27:55  gwyn
 * detach daemon from console process group if SYS3
 * 
 * Revision 1.9  89/02/23  05:50:05  gwyn
 * revised to use prodfile mtime instead of atime; fixed bugs
 * 
 * Revision 1.8  88/03/12  05:18:52  gwyn
 * *** empty log message ***
 * 
 * Revision 1.7  86/09/18  01:26:07  root
 * Changed from from using _NFILE to getdtablesize to determine
 * number of file descriptors (required for 4.3BSD), old code is
 * #ifdef'ed on _NFILE.
 * 
 * Revision 1.6  85/02/14  21:05:51  dpk
 * Combined DPW version with master.  Gained logging changes.
 * 
 * Revision 1.5  84/11/20  16:57:33  dpk
 * Added alarm(0) to termcatch to make system more resilient to
 * poor signal implementations causing race conditions. (SYSV)
 * 
 * Revision 1.4  85/02/02  21:46:49  dpk
 * Added code to support changing log files
 * 
 * Revision 1.3  84/10/17  01:40:04  dpk
 * Added support for RP_DELAY
 * 
 * Revision 1.2  84/10/02  13:24:26  dpk
 * New DPW versions
 * 
 * Revision 1.2  84/02/22  01:47:56  dpk
 * Added distinctive RCS header
 * 
 */
#ifndef lint
static char RCSid[] = "@(#)$Header: dmain.c,v 1.10 89/02/23 13:27:55 gwyn Exp $";
#endif

/*
 *		M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *
 *				  D M A I N . C
 *
 *			     Douglas P. Kingston III
 *				Michael John Muuss
 *
 *				 Revision History
 *				 ================
 *	13 Sep 82  DPK	Added signal trapping, changed checkdelreq() to
 *			checkdelayed(), added code for Waveoff's.
 *
 *	9 Nov 82   DPK  Changed init() so delayed qlist is initialized
 *
 *	17 Nov 82  DPK	Modified alarmcatch to do an alarm(0)
 *			Added queueing system shutdown code.
 *
 *	28 Nov 82  DPK	Added call to recover().  Force call to rdconf()
 *			early in init().  Added calls to unlock(&Dae_lock);
 *
 *	3  Mar 83  DAW	Added ifdef SDB so daemon could run under sdb.
 *
 *	11 Apr 83  DPK	Added umask(0) after fork of daemon in init().
 */

#include <stdio.h>
#include "queue.h"		/* Includes types.h, dir.h, qconf.h */
#include <signal.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/file.h>

extern	char *Qcntrldir;
extern	char *Qconffile;
extern	char *Seqfile;
extern	char *Prodfile;
extern	char *Newprefix;
extern	char *Modprefix;
extern	char *Logfile;
extern	lock_t Dae_lock;
extern	struct paramtab paramtab[];

struct	request *reqhead	= NULL;
struct	delreq *delhead		= NULL;
struct	device *devhead		= NULL;
struct	queue *quehead		= NULL;
int	delayupdate		= 0; 
int	dieflag			= 0;
time_t	confmod			= 0L;	/* modtime of configuration file */
time_t	newmodtime		= 0L;	/* modtime of new request dir */
time_t	modmodtime		= 0L;	/* modtime of modification dir */
time_t	prodtime		= 0L;	/* access time of prodfile */
ino_t	logino			= 0;	/* inode of logfile */
dev_t	logdev			= 0;	/* device of logfile */
int	prodfd;

/*
 *	MDQS Parameter variables
 */
int	Scanwait	= SCANWAIT;	/* Seconds to wait in main loop */
int	Procwait	= PROCWAIT;	/* Seconds to wait for a dead child */
int	Openwait	= OPENWAIT;	/* Seconds to wait on device open failure */
int	Delaytime	= DELAYTIME;	/* Seconds to delay a request on netfail */
int	Maxfailures	= 0;		/* Number of open failures tolerated */
int	Debug		= 0;		/* Debugging level */
int	Waveoff		= 0;		/* Set if no more room for requests */
char	*MDQSuser = "mdqs";		/* Mail from queueing system will come
					 * from this address.  It should be
					 * be a valid mail address.
					 */
char	*Sysmgr = "mdqs";		/* Orphaned notices will go to
					 * this address.  It should be a
					 * valid mail address.
					 */

char	**origargv;	/* Used in devstart() to change arg0 for ps(1). */
struct	request *makereq();
char	*strdup();

char	obuf[BUFSIZ];	/* For buffering Daemon prints */

/*
 *			M A I N
 *
 *	The sequenceing for events in the daemon is
 * controlled here in the mainline.
 */

main( argc, argv )
int	argc;
char	**argv;
{
	struct	stat	statbuf;	/* for stat()'s */

	/*
	 * Detach ourselves, Build internal tables, Go to Qcntrldir
	 */
	init( argv );

	/*
	 *	MAIN  LOOP
	 */
	while( 1 )  {
		if (fstat (prodfd, &statbuf) < 0) {
			complain ("Error fstatting %s (%s)", Prodfile,
				  sys_errlist[errno]);
		} else if (statbuf.st_mtime > prodtime) {
debug ("Prodfile updated.");
			prodtime = statbuf.st_mtime;
			statfiles();
			devcheck();
		}

		/*
		 * Check delayed requeses for recent activation
		 */
		checkdelayed();

		/*
		 *  Collect dead children, free devices for new tasks
		 */
		pwait();

		/*
		 *  For all inactive devices with requests pending,
		 *  initiate a queue device manager process.
		 */
		scandevs();

		qlists();
		prreq();

		/*
		 *  Delay for a while
		 */
		sleep( Scanwait );
	}
}

/*
 *			I N I T
 *
 *	Initialize the freelist for requests and delayed requests.
 */

init( argv )
register char	**argv;
{
	register int	fd;
	int	pid, prodid;
	int	alarmcatch();
	int	termcatch();
	struct	stat statbuf;
	register struct paramtab *pp;
	char 	*buf, linebuf[LINESIZ];

	/*
	 * Put ourselves into the queue directory
	 */
	if( chdir( Qcntrldir ) < 0 )  {
		complain("Can not CHDIR to %s", Qcntrldir );
		exit(10);
	}

	/*
	 *	Detach the daemon by forking and having parent exit.
	 *	The real daemon is actually the child.
	 */
#ifdef SDB
	printf( "debug: fake fork\n" );
#else
	if(( pid = fork()) < 0 )
		bomb( "Cannot detach daemon (%s)", sys_errlist[errno] );
	else if( pid != 0 )
		exit( 0 );	/* Parent */
#endif
#ifdef SYS3
	setpgrp();		/* detach from console */
#endif

	/* Save argv in a reachable place for devstart() */
	origargv = argv;

	/*
	 *	Trap unwanted signals
	 */
	signal( SIGHUP, SIG_IGN );
	signal( SIGINT, SIG_IGN );
	signal( SIGQUIT, SIG_IGN );
	signal( SIGALRM, alarmcatch );
	signal( SIGTERM, termcatch );

	umask (0);
	
	/*  Close all unnecessary filedescriptors  */
#ifndef _NFILE
	for( fd = getdtablesize()-1; fd > 2; fd-- )
#else
	for( fd = _NFILE-1; fd > 2; fd-- )
#endif
		close( fd );

	/*  Make sure there is only ONE daemon running at a time  */
	if( lock( &Dae_lock, NOWAIT ) == -1)
		bomb("There already appears to be a daemon.");

	setbuf( stderr, obuf );

	for( pp = paramtab; pp->paramstr != 0; pp++ )
		if( pp->vartype == STRPVAR && pp->parmp != NULL )
			*pp->parmp = strdup( *pp->parmp );

	/*
	 *  Process configuration file (forms, devs, etc)
	 */
	if( stat( Qconffile, &statbuf) < 0 ) {
		bomb("init(): Unable to access %s", Qconffile);
	} else {
		confmod = statbuf.st_mtime;
		rdconf ();
	}
	loginit();	/* Must do this after reading config file */
	if (stat (Logfile, &statbuf) == 0) {
		logino = statbuf.st_ino;
		logdev = statbuf.st_dev;
	}

	/*
	 *	Make sure the Sequence file is present and read/write
	 *	available.  If not, create one.
	 */
	if( access( Seqfile, RACCESS|WACCESS ) < 0 ) {
		complain( "Access error on %s (%s)",
			Seqfile, sys_errlist[errno] );
		complain( "Creating new sequence file..." );
		unlink( Seqfile );
		if(close( creat( Seqfile, 0666 )) < 0)
			bomb( "Could not creat sequence file.  Help." );
		chmod (Seqfile, 0666);
	}

	/*
	 *	Make sure the Prodfile is present and writable.
	 *	Added code to write the pid of the daemon into
	 *	the Prodfile.  Seems not all systems update the
	 *	access time when no. of bytes read is 0.  This
	 *	has been changed to use file modification anyway.
	 */


	if( access( Prodfile, RACCESS|WACCESS ) < 0 ) {
		if (errno == ENOENT) {
			complain( "Creating new prodfile..." );
			if ((prodfd = creat( Prodfile, 0666 )) < 0)
				bomb( "Could not creat prodfile.  Help." );
			else {
				close(prodfd);
				buf = linebuf;
				prodid = getpid();
				prodfd = open(Prodfile, O_WRONLY);
				sprintf(buf, "%d \n", prodid);
				write(prodfd, buf, strlen(buf));
				close(prodfd);
				chmod (Prodfile, 0666);
			}
		} else {
			bomb( "Access error on %s (%s).  Help.",
				Prodfile, sys_errlist[errno] );
		}
	}
	if ((prodfd = open (Prodfile, 0)) < 0) {
		bomb( "Open error on %s (%s).  Help.",
			Prodfile, sys_errlist[errno] );
	}

	/*  Force queue files recovery  */
	recover();
	scandevs();

	/*
	 *	Force an initially empty qlist file for delayed
	 *	requests to be made on the first called to qlists().
	 */
	delayupdate = 1;
}

statfiles() {
	struct	stat	statbuf;	/* for stat()'s */

	/*
	 *  Process any changes in configuration (forms, devs, etc)
	 */
	if( stat( Qconffile, &statbuf) < 0 ) {
		complain("Unable to access %s", Qconffile);
	} else if( statbuf.st_mtime > confmod )  {
		confmod = statbuf.st_mtime;
		rdconf ();
		devcheck ();
	}

	/*
	 * Process any changes in modifications directory
	 */
	if( stat( Modprefix, &statbuf ) < 0 ) {
		complain("Unable to access %s", Modprefix);
	} else if( statbuf.st_mtime > modmodtime )  {
		modmodtime = statbuf.st_mtime;
debug ("main: read modifications");
		rdmod();
	}

	/*
	 * Process any changes in Queue directory
	 */
	if( stat( Newprefix, &statbuf ) < 0 ) {
		complain("Unable to access %s", Newprefix);
	} else if( !Waveoff && statbuf.st_mtime > newmodtime )  {
		rdnew();
		if( !Waveoff )
			newmodtime = statbuf.st_mtime;
	}

	/*
	 * If logfile has changed, reopen the log file
	 */
	if( stat( Logfile, &statbuf ) == 0 ) {
		if (logino != statbuf.st_ino ||
		    logdev != statbuf.st_dev)
			loginit();
	} else {
		/* NO log file exists now ... */
		logino = logdev = 0;
	}
}

alarmcatch() {
	signal (SIGALRM, alarmcatch);
	alarm( 0 );
}

termcatch() {
	/*
	 *  Queueing system shutdown sequence
	 */
	signal (SIGTERM, SIG_IGN);
	alarm(0);
debug("Termination signal received...");
	termall();
	unlock( &Dae_lock );
	complain("Queueing system shutdown.");
	exit(0);
}

/*
 *			D E B U G
 *	Controllable complain().
 */
debug( fmt, a, b, c, d )
char	*fmt, *a, *b, *c, *d;
{
	if( Debug )
		comment( fmt, a, b, c, d );
}

/*
 *			C O M P L A I N
 *
 *	This function is designed to log an error message
 *  on a pre-specified device (presumably the system console).
 *  If the system assigns a controlling device when you open
 *  the console, define NOCONTROL and a child will do the open
 *  to keep the daemon from gaining a controlling tty.
 */
complain( a, b, c, d, e )
{
	comment( a, b, c, d, e );
	log (a, b, c, d, e);
}

comment( a, b, c, d, e )
{
	fputs( "MDQS: ", stderr );
	fprintf( stderr, a, b, c, d, e );
	fputc( '\n', stderr );
	fflush( stderr );
}

err (a, b, c, d, e)
char *a, *b, *c, *d, *e;
{
	complain (a, b, c, d, e);
}

/*
 *			    B O M B
 *
 *	Like complain() but for fatal daemon errors.
 */
bomb( a, b, c, d, e )  {
	fputs( "MDQS: ", stderr );
	fprintf( stderr, a, b, c, d, e );
	fputc( '\n', stderr );
#ifdef COREDUMP
	fputs( "MDQS: This is a FATAL error.  Dumping core...\n", stderr );
#else
	fputs( "MDQS: This is a FATAL error.\n", stderr );
#endif
	fflush( stderr );
	log (a, b, c, d, e);
	unlock( &Dae_lock );
#ifdef COREDUMP
	abort ();
#endif
	exit (99);
}

/*
 *			P R R E Q
 */
prreq()  {
	register struct device *dp;
	register struct queue *qp;
	register struct request *rp;

	if( Debug < 5 )
		return;

	fprintf( stderr, "\nDevices:\n" );
	for (dp = devhead; dp != NULL; dp = dp->dv_nextdev) {
		fprintf( stderr,
			"    %s: device %s, forms %s, status %d, failed %d\n",
			dp->dv_dname, dp->dv_device, dp->dv_forms,
			dp->dv_status, dp->dv_failed );
		if( dp->dv_pid != 0 )
			fprintf( stderr, "	Active:  %.14s (%d)\n",
				dp->dv_actreq->r_name, dp->dv_pid);
	}
	fprintf( stderr, "Queues:\n" );
	for (qp = quehead; qp != NULL; qp = qp->q_nextq) {
		fprintf( stderr, "    %s:\n", qp->q_name);
		rp = qp->q_readyreq;
		for(; rp != NULL; rp = rp->r_nextreq) {
			fprintf( stderr, "	Waiting: %.14s, form %s, uid %d, priority %d",
			    rp->r_name, rp->r_forms, rp->r_uid, rp->r_pri );
			if( rp->r_status & SKILL )
				fputs( "  Killed?", stderr );
			if( rp->r_status & SHOLD )
				fputs( "  Held", stderr );
			fputc( '\n', stderr );
		}
	}
	fflush( stderr );
}
